// Package transport provides Thrift TTransport and TServerTransport
// implementations for use on mac/linux (TSocket/TServerSocket) and Windows
// (custom named pipe implementation).
package transport
